"""The Mirascope LiteLLM Module."""

from .call import litellm_call
from .call import litellm_call as call
from .call_params import LiteLLMCallParams
from .call_response import LiteLLMCallResponse
from .call_response_chunk import LiteLLMCallResponseChunk
from .dynamic_config import LiteLLMDynamicConfig
from .tool import LiteLLMTool

__all__ = [
    "call",
    "LiteLLMDynamicConfig",
    "LiteLLMCallParams",
    "LiteLLMCallResponse",
    "LiteLLMCallResponseChunk",
    "LiteLLMTool",
    "litellm_call",
]



"""The `litellm_call` decorator for functions as LLM calls."""

from openai.types.chat import (
    ChatCompletionAssistantMessageParam,
    ChatCompletionMessageParam,
    ChatCompletionUserMessageParam,
)

from ..base import call_factory
from ..base._stream import BaseStream
from ._utils import (
    calculate_cost,
    get_json_output,
    handle_stream,
    handle_stream_async,
    setup_call,
)
from .call_params import LiteLLMCallParams
from .call_response import LiteLLMCallResponse
from .call_response_chunk import LiteLLMCallResponseChunk
from .dynamic_config import LiteLLMDynamicConfig
from .tool import LiteLLMTool


class LiteLLMStream(
    BaseStream[
        LiteLLMCallResponse,
        LiteLLMCallResponseChunk,
        ChatCompletionUserMessageParam,
        ChatCompletionAssistantMessageParam,
        ChatCompletionMessageParam,
        LiteLLMTool,
        LiteLLMDynamicConfig,
        LiteLLMCallParams,
    ]
):
    _provider = "litellm"


litellm_call = call_factory(
    TCallResponse=LiteLLMCallResponse,
    TCallResponseChunk=LiteLLMCallResponseChunk,
    TDynamicConfig=LiteLLMDynamicConfig,
    TMessageParamType=ChatCompletionAssistantMessageParam,
    TToolType=LiteLLMTool,
    TStream=LiteLLMStream,
    TCallParams=LiteLLMCallParams,
    default_call_params=LiteLLMCallParams(),
    setup_call=setup_call,  # type: ignore
    get_json_output=get_json_output,
    handle_stream=handle_stream,
    handle_stream_async=handle_stream_async,
    calculate_cost=calculate_cost,
    provider="litellm",
)
'''A decorator for calling the LiteLLM API with a typed function.

This decorator is used to wrap a typed function that calls the LiteLLM API. It parses
the docstring of the wrapped function as the messages array and templates the input
arguments for the function into each message's template.

Example:

```python
@litellm_call(model="gpt-4o")
def recommend_book(genre: str):
    """Recommend a {genre} book."""

response = recommend_book("fantasy")
```

Args:
    model: The LiteLLM model to use in the API call.
    stream: Whether to stream the response from the API call.
    tools: The tools to use in the LiteLLM API call.
    **call_params: The `LiteLLMCallParams` call parameters to use in the API call.

Returns:
    The decorator for turning a typed function into an LiteLLM API call.
'''



"""Calculate the cost of a LiteLLM API call."""

from litellm.batches.main import ModelResponse


def calculate_cost(response: ModelResponse, model: str) -> float | None:
    """Calculate the cost of a Gemini API call."""
    return None



"""This module contains the setup_call function for OpenAI tools."""

import inspect
from typing import Any, Awaitable, Callable

from litellm import acompletion, completion
from openai import AsyncAzureOpenAI, AsyncOpenAI, AzureOpenAI, OpenAI
from openai.types.chat import ChatCompletion, ChatCompletionMessageParam

from ...base import BaseTool, _utils
from ..call_params import LiteLLMCallParams
from ..dynamic_config import LiteLLMDynamicConfig
from ..tool import LiteLLMTool


def setup_call(
    *,
    model: str,
    client: OpenAI | AsyncOpenAI | AzureOpenAI | AsyncAzureOpenAI | None,
    fn: Callable[..., LiteLLMDynamicConfig | Awaitable[LiteLLMDynamicConfig]],
    fn_args: dict[str, Any],
    dynamic_config: LiteLLMDynamicConfig,
    tools: list[type[BaseTool] | Callable] | None,
    json_mode: bool,
    call_params: LiteLLMCallParams,
    extract: bool,
) -> tuple[
    Callable[..., ChatCompletion] | Callable[..., Awaitable[ChatCompletion]],
    str,
    list[ChatCompletionMessageParam],
    list[type[LiteLLMTool]] | None,
    dict[str, Any],
]:
    prompt_template, messages, tool_types, call_kwargs = _utils.setup_call(
        fn, fn_args, dynamic_config, tools, LiteLLMTool, call_params
    )
    if client is not None:
        raise ValueError(
            "Mirascope's LiteLLM integration doesn't currently support custom clients."
        )
    create = acompletion if inspect.iscoroutinefunction(fn) else completion
    call_kwargs |= {"model": model, "messages": messages}
    if json_mode:
        call_kwargs["response_format"] = {"type": "json_object"}
        messages[-1]["content"] += _utils.json_mode_content(
            tool_types[0] if tool_types else None
        )
        call_kwargs.pop("tools", None)
    elif extract:
        assert tool_types, "At least one tool must be provided for extraction."
        call_kwargs["tool_choice"] = "required"

    return create, prompt_template, messages, tool_types, call_kwargs  # type: ignore



"""This module contains the `LiteLLMCallResponse` class."""

from typing import Any

from litellm.batches.main import ModelResponse
from openai.types.chat import (
    ChatCompletionAssistantMessageParam,
    ChatCompletionMessage,
    ChatCompletionMessageParam,
    ChatCompletionMessageToolCall,
    ChatCompletionToolMessageParam,
    ChatCompletionUserMessageParam,
)
from openai.types.chat.chat_completion import Choice
from pydantic import computed_field

from ..base import BaseCallResponse
from .call_params import LiteLLMCallParams
from .dynamic_config import LiteLLMDynamicConfig
from .tool import LiteLLMTool


class LiteLLMCallResponse(
    BaseCallResponse[
        ModelResponse,
        LiteLLMTool,
        LiteLLMDynamicConfig,
        ChatCompletionMessageParam,
        LiteLLMCallParams,
        ChatCompletionUserMessageParam,
    ]
):
    '''A convenience wrapper around the LiteLLM `ChatCompletion` response.

    When calling the LiteLLM API using a function decorated with `openai_call`, the
    response will be an `LiteLLMCallResponse` instance with properties that allow for
    more convenience access to commonly used attributes.

    Example:

    ```python
    from mirascope.core.openai import openai_call

    @openai_call(model="gpt-4o")
    def recommend_book(genre: str):
        """Recommend a {genre} book."""

    response = recommend_book("fantasy")  # response is an `LiteLLMCallResponse` instance
    print(response.content)
    #> Sure! I would recommend...
    ```
    '''

    _provider = "openai"

    @computed_field
    @property
    def message_param(self) -> ChatCompletionAssistantMessageParam:
        """Returns the assistants's response as a message parameter."""
        return self.message.model_dump(exclude={"function_call"})  # type: ignore

    @property
    def choices(self) -> list[Choice]:
        """Returns the array of chat completion choices."""
        return self.response.choices  # type: ignore

    @property
    def choice(self) -> Choice:
        """Returns the 0th choice."""
        return self.choices[0]

    @property
    def message(self) -> ChatCompletionMessage:
        """Returns the message of the chat completion for the 0th choice."""
        return self.choice.message

    @property
    def content(self) -> str:
        """Returns the content of the chat completion for the 0th choice."""
        return self.message.content if self.message.content is not None else ""

    @property
    def model(self) -> str | None:
        """Returns the name of the response model."""
        return self.response.model

    @property
    def id(self) -> str:
        """Returns the id of the response."""
        return self.response.id

    @property
    def finish_reasons(self) -> list[str]:
        """Returns the finish reasons of the response."""
        return [str(choice.finish_reason) for choice in self.response.choices]

    @property
    def tool_calls(self) -> list[ChatCompletionMessageToolCall] | None:
        """Returns the tool calls for the 0th choice message."""
        return self.message.tool_calls

    @computed_field
    @property
    def tools(self) -> list[LiteLLMTool] | None:
        """Returns any available tool calls as their `LiteLLMTool` definition.

        Raises:
            ValidationError: if a tool call doesn't match the tool's schema.
        """
        if not self.tool_types or not self.tool_calls:
            return None

        extracted_tools = []
        for tool_call in self.tool_calls:
            for tool_type in self.tool_types:
                if tool_call.function.name == tool_type._name():
                    extracted_tools.append(tool_type.from_tool_call(tool_call))
                    break

        return extracted_tools

    @computed_field
    @property
    def tool(self) -> LiteLLMTool | None:
        """Returns the 0th tool for the 0th choice message.

        Raises:
            ValidationError: if the tool call doesn't match the tool's schema.
        """
        if tools := self.tools:
            return tools[0]
        return None

    @classmethod
    def tool_message_params(
        cls, tools_and_outputs: list[tuple[LiteLLMTool, str]]
    ) -> list[ChatCompletionToolMessageParam]:
        """Returns the tool message parameters for tool call results."""
        return [
            ChatCompletionToolMessageParam(
                role="tool",
                content=output,
                tool_call_id=tool.tool_call.id,
                name=tool._name(),  # type: ignore
            )
            for tool, output in tools_and_outputs
        ]

    @property
    def usage(self) -> Any | None:
        """Returns the usage of the chat completion."""
        return self.response.usage  # type: ignore

    @property
    def input_tokens(self) -> int | None:
        """Returns the number of input tokens."""
        return self.usage.prompt_tokens if self.usage else None

    @property
    def output_tokens(self) -> int | None:
        """Returns the number of output tokens."""
        return self.usage.completion_tokens if self.usage else None



"""This module defines the base class for structured streams."""

import inspect
from collections.abc import AsyncGenerator, Generator
from functools import wraps
from re import T
from typing import (
    AsyncIterable,
    Awaitable,
    Callable,
    Generic,
    Iterable,
    ParamSpec,
    TypeVar,
    overload,
)

from pydantic import BaseModel

from ._utils import (
    BaseType,
    GetJsonOutput,
    SetupCall,
    extract_tool_return,
    get_fn_args,
    setup_extract_tool,
)
from .call_params import BaseCallParams
from .call_response_chunk import BaseCallResponseChunk
from .dynamic_config import BaseDynamicConfig
from .tool import BaseTool

_BaseCallResponseChunkT = TypeVar(
    "_BaseCallResponseChunkT", bound=BaseCallResponseChunk
)
_AssistantMessageParamT = TypeVar("_AssistantMessageParamT")
_BaseToolT = TypeVar("_BaseToolT", bound=BaseTool)
_BaseCallParamsT = TypeVar("_BaseCallParamsT", bound=BaseCallParams)
_BaseDynamicConfigT = TypeVar("_BaseDynamicConfigT", bound=BaseDynamicConfig)
_ResponseModelT = TypeVar("_ResponseModelT", bound=BaseModel | BaseType)


class BaseStructuredStream(Generic[_BaseCallResponseChunkT, _ResponseModelT]):
    """A base class for streaming structured outputs from LLMs."""

    stream: (
        Generator[_BaseCallResponseChunkT, None, None]
        | AsyncGenerator[_BaseCallResponseChunkT, None]
    )
    response_model: type[_ResponseModelT]
    json_mode: bool
    get_json_output: GetJsonOutput[_BaseCallResponseChunkT]

    def __init__(
        self,
        stream: Generator[_BaseCallResponseChunkT, None, None]
        | AsyncGenerator[_BaseCallResponseChunkT, None],
        *,
        response_model: type[_ResponseModelT],
        json_mode: bool,
        get_json_output: Callable,
    ):
        """Initializes an instance of `BaseStructuredStream`."""
        self.stream = stream
        self.response_model = response_model
        self.json_mode = json_mode
        self.get_json_output = get_json_output

    def __iter__(self) -> Generator[_ResponseModelT, None, None]:
        """Iterates over the stream and extracts structured outputs."""
        assert isinstance(
            self.stream, Generator
        ), "Stream must be a generator for __iter__"
        json_output = ""
        for chunk in self.stream:
            json_output += self.get_json_output(chunk, self.json_mode)
            if json_output and json_output[0] != "{":
                try:
                    json_start = json_output.index("{")
                    json_output = json_output[json_start:]
                except ValueError:
                    json_output = ""
            if chunk.model is not None:
                self.model = chunk.model
            if json_output:
                yield extract_tool_return(self.response_model, json_output, True)
        if json_output:
            json_output = json_output[: json_output.rfind("}") + 1]
        yield extract_tool_return(self.response_model, json_output, False)

    def __aiter__(self) -> AsyncGenerator[_ResponseModelT, None]:
        """Iterates over the stream and extracts structured outputs."""

        async def generator():
            assert isinstance(
                self.stream, AsyncGenerator
            ), "Stream must be an async generator for __aiter__"
            json_output = ""
            async for chunk in self.stream:
                json_output += self.get_json_output(chunk, self.json_mode)
                if json_output and json_output[0] != "{":
                    try:
                        json_start = json_output.index("{")
                        json_output = json_output[json_start:]
                    except ValueError:
                        json_output = ""
                if chunk.model is not None:
                    self.model = chunk.model
                if json_output:
                    yield extract_tool_return(self.response_model, json_output, True)
            if json_output:
                json_output = json_output[: json_output.rfind("}") + 1]
            yield extract_tool_return(self.response_model, json_output, False)

        return generator()


_BaseDynamicConfigT = TypeVar("_BaseDynamicConfigT", bound=BaseDynamicConfig)
_BaseClientT = TypeVar("_BaseClientT", bound=object)
_ResponseT = TypeVar("_ResponseT")
_ResponseChunkT = TypeVar("_ResponseChunkT")
_P = ParamSpec("_P")


def structured_stream_factory(
    *,
    TCallResponseChunk: type[_BaseCallResponseChunkT],
    TToolType: type[_BaseToolT],
    setup_call: SetupCall[
        _BaseClientT,
        _BaseDynamicConfigT,
        _BaseCallParamsT,
        _ResponseT,
        _ResponseChunkT,
        _BaseToolT,
    ],
    get_json_output: GetJsonOutput[_BaseCallResponseChunkT],
):
    @overload
    def decorator(
        fn: Callable[_P, _BaseDynamicConfigT],
        model: str,
        response_model: type[_ResponseModelT],
        json_mode: bool,
        client: _BaseClientT | None,
        call_params: _BaseCallParamsT,
    ) -> Callable[
        _P,
        Iterable[_ResponseModelT],
    ]: ...  # pragma: no cover

    @overload
    def decorator(
        fn: Callable[_P, Awaitable[_BaseDynamicConfigT]],
        model: str,
        response_model: type[_ResponseModelT],
        json_mode: bool,
        client: _BaseClientT | None,
        call_params: _BaseCallParamsT,
    ) -> Callable[
        _P,
        Awaitable[AsyncIterable[_ResponseModelT]],
    ]: ...  # pragma: no cover

    def decorator(
        fn: Callable[_P, _BaseDynamicConfigT | Awaitable[_BaseDynamicConfigT]],
        model: str,
        response_model: type[_ResponseModelT],
        json_mode: bool,
        client: _BaseClientT | None,
        call_params: _BaseCallParamsT,
    ) -> Callable[
        _P,
        Iterable[_ResponseModelT] | Awaitable[AsyncIterable[_ResponseModelT]],
    ]:
        tool = setup_extract_tool(response_model, TToolType)
        is_async = inspect.iscoroutinefunction(fn)

        @wraps(fn)
        def inner(*args: _P.args, **kwargs: _P.kwargs) -> Iterable[_ResponseModelT]:
            assert SetupCall.fn_is_sync(fn)
            fn_args = get_fn_args(fn, args, kwargs)
            dynamic_config = fn(*args, **kwargs)
            create, _, _, _, call_kwargs = setup_call(
                model=model,
                client=client,
                fn=fn,
                fn_args=fn_args,
                dynamic_config=dynamic_config,
                tools=[tool],
                json_mode=json_mode,
                call_params=call_params,
                extract=True,
            )
            stream = create(stream=True, **call_kwargs)
            return BaseStructuredStream[_BaseCallResponseChunkT, _ResponseModelT](
                (TCallResponseChunk(chunk=chunk) for chunk in stream),
                response_model=response_model,
                json_mode=json_mode,
                get_json_output=get_json_output,
            )

        async def inner_async(
            *args: _P.args, **kwargs: _P.kwargs
        ) -> AsyncIterable[_ResponseModelT]:
            assert SetupCall.fn_is_async(fn)
            fn_args = get_fn_args(fn, args, kwargs)
            dynamic_config = await fn(*args, **kwargs)
            create, _, _, _, call_kwargs = setup_call(
                model=model,
                client=client,
                fn=fn,
                fn_args=fn_args,
                dynamic_config=dynamic_config,
                tools=[tool],
                json_mode=json_mode,
                call_params=call_params,
                extract=True,
            )
            stream = await create(stream=True, **call_kwargs)
            return BaseStructuredStream[_BaseCallResponseChunkT, _ResponseModelT](
                (TCallResponseChunk(chunk=chunk) async for chunk in stream),
                response_model=response_model,
                json_mode=json_mode,
                get_json_output=get_json_output,
            )

        return inner_async if is_async else inner

    return decorator



"""The `extract_factory` method for generating provider specific create decorators."""

import datetime
import inspect
from functools import wraps
from typing import Awaitable, Callable, ParamSpec, TypeVar, overload

from pydantic import BaseModel, ValidationError

from ._utils import (
    BaseType,
    CalculateCost,
    GetJsonOutput,
    SetupCall,
    extract_tool_return,
    get_fn_args,
    get_metadata,
    setup_extract_tool,
)
from .call_params import BaseCallParams
from .call_response import BaseCallResponse
from .dynamic_config import BaseDynamicConfig
from .tool import BaseTool

_BaseCallResponseT = TypeVar("_BaseCallResponseT", bound=BaseCallResponse)
_BaseClientT = TypeVar("_BaseClientT", bound=object)
_BaseDynamicConfigT = TypeVar("_BaseDynamicConfigT", bound=BaseDynamicConfig)
_BaseCallParamsT = TypeVar("_BaseCallParamsT", bound=BaseCallParams)
_ResponseT = TypeVar("_ResponseT")
_ResponseChunkT = TypeVar("_ResponseChunkT")
_BaseToolT = TypeVar("_BaseToolT", bound=BaseTool)
_ResponseModelT = TypeVar("_ResponseModelT", bound=BaseModel | BaseType)
_P = ParamSpec("_P")


def extract_factory(
    *,
    TCallResponse: type[_BaseCallResponseT],
    TToolType: type[BaseTool],
    setup_call: SetupCall[
        _BaseClientT,
        _BaseDynamicConfigT,
        _BaseCallParamsT,
        _ResponseT,
        _ResponseChunkT,
        _BaseToolT,
    ],
    get_json_output: GetJsonOutput[_BaseCallResponseT],
    calculate_cost: CalculateCost[_ResponseT],
):
    """Returns the wrapped function with the provider specific interfaces."""

    @overload
    def decorator(
        fn: Callable[_P, _BaseDynamicConfigT],
        model: str,
        response_model: type[_ResponseModelT],
        json_mode: bool,
        client: _BaseClientT | None,
        call_params: _BaseCallParamsT,
    ) -> Callable[_P, _ResponseModelT]: ...  # pragma: no cover

    @overload
    def decorator(
        fn: Callable[_P, Awaitable[_BaseDynamicConfigT]],
        model: str,
        response_model: type[_ResponseModelT],
        json_mode: bool,
        client: _BaseClientT | None,
        call_params: _BaseCallParamsT,
    ) -> Callable[_P, Awaitable[_ResponseModelT]]: ...  # pragma: no cover

    def decorator(
        fn: Callable[_P, _BaseDynamicConfigT | Awaitable[_BaseDynamicConfigT]],
        model: str,
        response_model: type[_ResponseModelT],
        json_mode: bool,
        client: _BaseClientT | None,
        call_params: _BaseCallParamsT,
    ) -> Callable[_P, _ResponseModelT | Awaitable[_ResponseModelT]]:
        tool = setup_extract_tool(response_model, TToolType)
        is_async = inspect.iscoroutinefunction(fn)

        @wraps(fn)
        def inner(*args: _P.args, **kwargs: _P.kwargs) -> _ResponseModelT:
            assert SetupCall.fn_is_sync(fn)
            fn_args = get_fn_args(fn, args, kwargs)
            dynamic_config = fn(*args, **kwargs)
            create, prompt_template, messages, tool_types, call_kwargs = setup_call(
                model=model,
                client=client,
                fn=fn,
                fn_args=fn_args,
                dynamic_config=dynamic_config,
                tools=[tool],
                json_mode=json_mode,
                call_params=call_params,
                extract=True,
            )
            start_time = datetime.datetime.now().timestamp() * 1000
            response = create(stream=False, **call_kwargs)
            end_time = datetime.datetime.now().timestamp() * 1000
            call_response = TCallResponse(
                metadata=get_metadata(fn, dynamic_config),
                response=response,
                tool_types=tool_types,
                prompt_template=prompt_template,
                fn_args=fn_args,
                dynamic_config=dynamic_config,
                messages=messages,
                call_params=call_kwargs,
                user_message_param=messages[-1]
                if messages[-1]["role"] == "user"
                else None,
                start_time=start_time,
                end_time=end_time,
                cost=calculate_cost(response, model),
            )
            json_output = get_json_output(call_response, json_mode)
            try:
                output = extract_tool_return(response_model, json_output, False)
            except ValidationError as e:
                e._response = call_response  # type: ignore
                raise e
            if isinstance(output, BaseModel):
                output._response = call_response  # type: ignore
            return output  # type: ignore

        @wraps(fn)
        async def inner_async(*args: _P.args, **kwargs: _P.kwargs) -> _ResponseModelT:
            assert SetupCall.fn_is_async(fn)
            fn_args = get_fn_args(fn, args, kwargs)
            dynamic_config = await fn(*args, **kwargs)
            create, prompt_template, messages, tool_types, call_kwargs = setup_call(
                model=model,
                client=client,
                fn=fn,
                fn_args=fn_args,
                dynamic_config=dynamic_config,
                tools=[tool],
                json_mode=json_mode,
                call_params=call_params,
                extract=True,
            )
            start_time = datetime.datetime.now().timestamp() * 1000
            response = await create(stream=False, **call_kwargs)
            end_time = datetime.datetime.now().timestamp() * 1000
            call_response = TCallResponse(
                metadata=get_metadata(fn, dynamic_config),
                response=response,
                tool_types=tool_types,
                prompt_template=prompt_template,
                fn_args=fn_args,
                dynamic_config=dynamic_config,
                messages=messages,
                call_params=call_kwargs,
                user_message_param=messages[-1]
                if messages[-1]["role"] == "user"
                else None,
                start_time=start_time,
                end_time=end_time,
                cost=calculate_cost(response, model),
            )
            json_output = get_json_output(call_response, json_mode)
            output = extract_tool_return(response_model, json_output, False)
            if isinstance(output, BaseModel):
                output._response = call_response  # type: ignore
            return output  # type: ignore

        return inner_async if is_async else inner

    return decorator



"""This module contains the `CohereCallResponseChunk` class."""

from typing import Literal

from cohere import (
    StreamedChatResponse_CitationGeneration,
    StreamedChatResponse_SearchQueriesGeneration,
    StreamedChatResponse_SearchResults,
    StreamedChatResponse_StreamEnd,
    StreamedChatResponse_StreamStart,
    StreamedChatResponse_ToolCallsGeneration,
)
from cohere.types import (
    ApiMetaBilledUnits,
    ChatCitation,
    ChatDocument,
    ChatSearchQuery,
    ChatSearchResult,
    NonStreamedChatResponse,
    StreamedChatResponse,
    StreamedChatResponse_TextGeneration,
    ToolCall,
)
from pydantic import SkipValidation

from ..base import BaseCallResponseChunk


class CohereCallResponseChunk(
    BaseCallResponseChunk[SkipValidation[StreamedChatResponse]]
):
    '''A convenience wrapper around the Cohere `ChatCompletionChunk` streamed chunks.

    When calling the Cohere API using a function decorated with `cohere_call` and
    `stream` set to `True`, the stream will contain `CohereResponseChunk` instances with
    properties that allow for more convenient access to commonly used attributes.

    Example:

    ```python
    from mirascope.core.cohere import cohere_call

    @cohere_call(model="gpt-4o", stream=True)
    def recommend_book(genre: str):
        """Recommend a {genre} book."""

    stream = recommend_book("fantasy")  # response is an `CohereStream`
    for chunk in stream:
        print(chunk.content, end="", flush=True)
    #> Sure! I would recommend...
    ```
    '''

    @property
    def event_type(
        self,
    ) -> Literal[
        "stream-start",
        "search-queries-generation",
        "search-results",
        "text-generation",
        "citation-generation",
        "tool-calls-generation",
        "stream-end",
        "tool-calls-chunk",
    ]:
        """Returns the type of the chunk."""
        return self.chunk.event_type

    @property
    def content(self) -> str:
        """Returns the content for the 0th choice delta."""
        if isinstance(self.chunk, StreamedChatResponse_TextGeneration):
            return self.chunk.text
        return ""

    @property
    def search_queries(self) -> list[ChatSearchQuery] | None:
        """Returns the search queries for search-query event type else None."""
        if isinstance(self.chunk, StreamedChatResponse_SearchQueriesGeneration):
            return self.chunk.search_queries  # type: ignore
        return None

    @property
    def search_results(self) -> list[ChatSearchResult] | None:
        """Returns the search results for search-results event type else None."""
        if isinstance(self.chunk, StreamedChatResponse_SearchResults):
            return self.chunk.search_results
        return None

    @property
    def documents(self) -> list[ChatDocument] | None:
        """Returns the documents for search-results event type else None."""
        if isinstance(self.chunk, StreamedChatResponse_SearchResults):
            return self.chunk.documents
        return None

    @property
    def citations(self) -> list[ChatCitation] | None:
        """Returns the citations for citation-generation event type else None."""
        if isinstance(self.chunk, StreamedChatResponse_CitationGeneration):
            return self.chunk.citations
        return None

    @property
    def model(self) -> str | None:
        """Returns the name of the response model.

        Cohere does not return model, so we return None
        """
        return None

    @property
    def id(self) -> str | None:
        """Returns the id of the response."""
        if isinstance(self.chunk, StreamedChatResponse_StreamStart):
            return self.chunk.generation_id
        return None

    @property
    def finish_reasons(self) -> list[str] | None:
        """Returns the finish reasons of the response."""
        if isinstance(self.chunk, StreamedChatResponse_StreamEnd):
            return [str(self.chunk.finish_reason)]
        return None

    @property
    def response(self) -> NonStreamedChatResponse | None:
        """Returns the full response for the stream-end event type else None."""
        if isinstance(self.chunk, StreamedChatResponse_StreamEnd):
            return self.chunk.response
        return None

    @property
    def tool_calls(self) -> list[ToolCall] | None:
        """Returns the partial tool calls for the 0th choice message."""
        if isinstance(self.chunk, StreamedChatResponse_ToolCallsGeneration):
            return self.chunk.tool_calls
        return None

    @property
    def usage(self) -> ApiMetaBilledUnits | None:
        """Returns the usage of the response."""
        if isinstance(self.chunk, StreamedChatResponse_StreamEnd):
            if self.chunk.response.meta:
                return self.chunk.response.meta.billed_units
        return None

    @property
    def input_tokens(self) -> float | None:
        """Returns the number of input tokens."""
        if self.usage:
            return self.usage.input_tokens
        return None

    @property
    def output_tokens(self) -> float | None:
        """Returns the number of output tokens."""
        if self.usage:
            return self.usage.output_tokens
        return None



"""Utility for pulling metadata from a call and merging with any dynamic metadata."""

from typing import Callable

from ..dynamic_config import BaseDynamicConfig
from ..metadata import Metadata


def get_metadata(fn: Callable, dynamic_config: BaseDynamicConfig) -> Metadata:
    """Get the metadata from the function and merge with any dynamic metadata."""
    metadata = fn.__annotations__.get("metadata", {})
    if dynamic_metadata := getattr(dynamic_config, "metadata", {}):
        metadata = metadata | dynamic_metadata
    return metadata



"""Get JSON output from a Gemini response."""

from ..call_response import GeminiCallResponse
from ..call_response_chunk import GeminiCallResponseChunk


def get_json_output(
    response: GeminiCallResponse | GeminiCallResponseChunk, json_mode: bool
) -> str:
    """Extracts the JSON output from a Gemini response."""
    if isinstance(response, GeminiCallResponse):
        if json_mode and response.content:
            return response.content
        elif tool_calls := response.tool_calls:
            return tool_calls[0].function_call.args
        else:
            raise ValueError("No tool call or JSON object found in response.")
    elif not json_mode:
        raise ValueError("Gemini only supports structured streaming in json mode.")
    return response.content



"""Mirascope x Logfire Integration."""

from contextlib import contextmanager
from typing import (
    Any,
    Awaitable,
    Callable,
    Generator,
    ParamSpec,
    TypeVar,
    overload,
)

import logfire
from pydantic import BaseModel

from ..core.base import BaseCallResponse
from ..core.base._stream import BaseStream
from .utils import middleware

_P = ParamSpec("_P")
_R = TypeVar("_R", bound=BaseCallResponse | BaseStream | BaseModel)


@overload
def with_logfire(fn: Callable[_P, _R]) -> Callable[_P, _R]: ...


@overload
def with_logfire(fn: Callable[_P, Awaitable[_R]]) -> Callable[_P, Awaitable[_R]]: ...


def with_logfire(
    fn: Callable[_P, _R] | Callable[_P, Awaitable[_R]],
) -> Callable[_P, _R] | Callable[_P, Awaitable[_R]]:
    """Wraps a Mirascope function with Logfire tracing."""

    @contextmanager
    def custom_context_manager() -> Generator[logfire.LogfireSpan, Any, None]:
        metadata = fn.__annotations__.get("metadata", {})
        tags = getattr(metadata, "tags", {})
        with logfire.with_settings(
            custom_scope_suffix="mirascope", tags=list(tags)
        ).span(fn.__name__) as logfire_span:
            yield logfire_span

    def get_call_response_span_data(result: BaseCallResponse) -> dict:
        return {
            "async": False,
            "call_params": result.call_params,
            "model": result.model,
            "provider": result._provider,
            "prompt_template": result.prompt_template,
            "template_variables": result.fn_args,
            "messages": result.messages,
            "response_data": result.response,
        }

    def get_stream_span_data(stream: BaseStream) -> dict:
        return {
            "messages": [stream.user_message_param],
            "call_params": stream.call_params,
            "model": stream.model,
            "provider": stream.provider,
            "prompt_template": stream.prompt_template,
            "template_variables": stream.fn_args,
            "output": {
                "cost": stream.cost,
                "content": stream.message_param.get("content", None),
            },
        }

    def get_tool_calls(result: BaseCallResponse) -> list[dict] | None:
        if tools := result.tools:
            tool_calls = [
                {
                    "function": {
                        "arguments": tool.model_dump_json(exclude={"tool_call"}),
                        "name": tool.name(),
                    }
                }
                for tool in tools
            ]
            return tool_calls
        return None

    def get_output(result: BaseCallResponse) -> dict[str, Any]:
        output: dict[str, Any] = {}
        if cost := result.cost:
            output["cost"] = cost
        if input_tokens := result.input_tokens:
            output["input_tokens"] = input_tokens
        if output_tokens := result.output_tokens:
            output["output_tokens"] = output_tokens
        if content := result.content:
            output["content"] = content
        return output

    def handle_call_response(
        result: BaseCallResponse, logfire_span: logfire.LogfireSpan | None
    ):
        if logfire_span is None:
            return
        output = get_output(result)
        span_data = get_call_response_span_data(result)
        span_data["async"] = False
        tool_calls = get_tool_calls(result)
        if tool_calls:
            output["tool_calls"] = tool_calls
        span_data["output"] = output
        logfire_span.set_attributes(span_data)

    def handle_stream(
        stream: BaseStream,
        logfire_span: logfire.LogfireSpan | None,
    ):
        span_data = get_stream_span_data(stream) | {"async": False}
        if logfire_span is not None:
            logfire_span.set_attributes(span_data)

    def handle_base_model(result: BaseModel, logfire_span: logfire.LogfireSpan | None):
        span_data: dict[str, Any] = {}
        output: dict[str, Any] = {}
        if result._response is not None:
            response: BaseCallResponse = result._response
            span_data |= get_call_response_span_data(response)
            output |= get_output(response)
        output["response_model"] = result
        span_data["async"] = False
        span_data["output"] = output
        logfire_span.set_attributes(span_data)

    async def handle_call_response_async(
        result: BaseCallResponse, logfire_span: logfire.LogfireSpan | None
    ):
        if logfire_span is None:
            return
        output = get_output(result)
        span_data = get_call_response_span_data(result)
        span_data["async"] = True
        tool_calls = get_tool_calls(result)
        if tool_calls:
            output["tool_calls"] = tool_calls
        span_data["output"] = output
        logfire_span.set_attributes(span_data)

    async def handle_stream_async(
        stream: BaseStream,
        logfire_span: logfire.LogfireSpan | None,
    ):
        span_data = get_stream_span_data(stream) | {"async": True}
        if logfire_span is not None:
            logfire_span.set_attributes(span_data)

    async def handle_base_model_async(
        result: BaseModel, logfire_span: logfire.LogfireSpan | None
    ):
        span_data: dict[str, Any] = {}
        output: dict[str, Any] = {}
        if result._response is not None:
            response: BaseCallResponse = result._response
            span_data |= get_call_response_span_data(response)
            output |= get_output(response)
        output["response_model"] = result
        span_data["async"] = True
        span_data["output"] = output
        logfire_span.set_attributes(span_data)

    return middleware(
        fn,
        custom_context_manager=custom_context_manager,
        handle_call_response=handle_call_response,
        handle_call_response_async=handle_call_response_async,
        handle_stream=handle_stream,
        handle_stream_async=handle_stream_async,
        handle_base_model=handle_base_model,
        handle_base_model_async=handle_base_model_async,
    )



"""Cohere utilities for decorator factories."""

from .calculate_cost import calculate_cost
from .get_json_output import get_json_output
from .handle_stream import handle_stream, handle_stream_async
from .setup_call import setup_call

__all__ = [
    "calculate_cost",
    "get_json_output",
    "handle_stream",
    "handle_stream_async",
    "setup_call",
]



"""Calculate the cost of a Gemini API call."""

from google.generativeai.types import GenerateContentResponse  # type: ignore


def calculate_cost(response: GenerateContentResponse, model: str) -> float | None:
    """Calculate the cost of a Gemini API call."""
    return None



"""This module contains the `MistralCallResponseChunk` class."""

from mistralai.models.chat_completion import (
    ChatCompletionResponseStreamChoice,
    ChatCompletionStreamResponse,
    DeltaMessage,
    ToolCall,
)
from mistralai.models.common import UsageInfo

from ..base import BaseCallResponseChunk


class MistralCallResponseChunk(BaseCallResponseChunk[ChatCompletionStreamResponse]):
    '''A convenience wrapper around the Mistral `ChatCompletionChunk` streamed chunks.

    When calling the Mistral API using a function decorated with `mistral_call` and
    `stream` set to `True`, the stream will contain `MistralResponseChunk` instances with
    properties that allow for more convenient access to commonly used attributes.

    Example:

    ```python
    from mirascope.core.mistral import mistral_call

    @mistral_call(model="gpt-4o", stream=True)
    def recommend_book(genre: str):
        """Recommend a {genre} book."""

    stream = recommend_book("fantasy")  # response is an `MistralStream`
    for chunk in stream:
        print(chunk.content, end="", flush=True)
    #> Sure! I would recommend...
    ```
    '''

    @property
    def choices(self) -> list[ChatCompletionResponseStreamChoice]:
        """Returns the array of chat completion choices."""
        return self.chunk.choices

    @property
    def choice(self) -> ChatCompletionResponseStreamChoice:
        """Returns the 0th choice."""
        return self.choices[0]

    @property
    def delta(self) -> DeltaMessage:
        """Returns the delta of the 0th choice."""
        return self.choice.delta

    @property
    def content(self) -> str:
        """Returns the content of the delta."""
        return self.delta.content if self.delta.content is not None else ""

    @property
    def model(self) -> str:
        """Returns the name of the response model."""
        return self.chunk.model

    @property
    def id(self) -> str:
        """Returns the id of the response."""
        return self.chunk.id

    @property
    def finish_reasons(self) -> list[str]:
        """Returns the finish reasons of the response."""
        return [
            choice.finish_reason if choice.finish_reason else ""
            for choice in self.choices
        ]

    @property
    def tool_calls(self) -> list[ToolCall] | None:
        """Returns the partial tool calls for the 0th choice message."""
        return self.delta.tool_calls

    @property
    def usage(self) -> UsageInfo | None:
        """Returns the usage of the chat completion."""
        return self.chunk.usage

    @property
    def input_tokens(self) -> int | None:
        """Returns the number of input tokens."""
        if self.usage:
            return self.usage.prompt_tokens
        return None

    @property
    def output_tokens(self) -> int | None:
        """Returns the number of output tokens."""
        if self.usage:
            return self.usage.completion_tokens
        return None



"""The Mirascope Core Functionality."""

from contextlib import suppress

from . import base
from .base import (
    BasePrompt,
    BaseTool,
    BaseToolKit,
    metadata,
    prompt_template,
    toolkit_tool,
)

with suppress(ImportError):
    from . import openai as openai

__all__ = [
    "base",
    "BasePrompt",
    "BaseTool",
    "BaseToolKit",
    "metadata",
    "openai",
    "prompt_template",
    "toolkit_tool",
]



"""Mirascope x Langfuse Integration."""

from typing import (
    Awaitable,
    Callable,
    ParamSpec,
    TypeVar,
    overload,
)

from langfuse.decorators import langfuse_context, observe
from langfuse.types import ModelUsage
from pydantic import BaseModel

from ..core.base import BaseCallResponse
from ..core.base._stream import BaseStream
from .utils import middleware

_P = ParamSpec("_P")
_R = TypeVar("_R", bound=BaseCallResponse | BaseStream | BaseModel)


@overload
def with_langfuse(fn: Callable[_P, _R]) -> Callable[_P, _R]: ...


@overload
def with_langfuse(fn: Callable[_P, Awaitable[_R]]) -> Callable[_P, Awaitable[_R]]: ...


def with_langfuse(
    fn: Callable[_P, _R] | Callable[_P, Awaitable[_R]],
) -> Callable[_P, _R] | Callable[_P, Awaitable[_R]]:
    """Wraps a Mirascope function with Langfuse tracing."""

    def get_call_response_observation(result: BaseCallResponse):
        return {
            "name": f"{fn.__name__} with {result.model}",
            "input": result.prompt_template,
            "metadata": result.response,
            "tags": result.tags,
            "model": result.model,
            "output": result.message_param.get("content", None),
        }

    def get_stream_observation(stream: BaseStream):
        return {
            "name": f"{fn.__name__} with {stream.model}",
            "input": stream.prompt_template,
            "tags": fn.__annotations__.get("tags", []),
            "model": stream.model,
            "output": stream.message_param.get("content", None),
        }

    def handle_call_response(result: BaseCallResponse):
        usage = ModelUsage(
            input=result.input_tokens,
            output=result.output_tokens,
            unit="TOKENS",
        )
        langfuse_context.update_current_observation(
            **get_call_response_observation(result),
            usage=usage,
        )

    def handle_stream(stream: BaseStream):
        # TODO: Re-add once Mirascope supports streaming usage
        # usage = ModelUsage(
        #     input=stream.input_tokens,
        #     output=stream.output_tokens,
        #     unit="TOKENS",
        # )
        langfuse_context.update_current_observation(
            **get_stream_observation(stream),
            # usage=usage,
        )

    def handle_base_model(result: BaseModel):
        if result._response is not None:
            response: BaseCallResponse = result._response
            handle_call_response(response)
        langfuse_context.update_current_observation(output=result)

    async def handle_call_response_async(result: BaseCallResponse):
        handle_call_response(result)

    async def handle_stream_async(stream: BaseStream):
        handle_stream(stream)

    async def handle_base_model_async(result: BaseModel):
        handle_base_model(result)

    return middleware(
        fn,
        custom_decorator=observe(
            name=fn.__name__,
            as_type="generation",
            capture_input=False,
            capture_output=False,
        ),
        handle_call_response=handle_call_response,
        handle_call_response_async=handle_call_response_async,
        handle_stream=handle_stream,
        handle_stream_async=handle_stream_async,
        handle_base_model=handle_base_model,
        handle_base_model_async=handle_base_model_async,
    )



"""This module contains the base classes for streaming responses from LLMs."""

import inspect
from abc import ABC
from collections.abc import AsyncGenerator, Generator
from functools import wraps
from typing import (
    Any,
    Awaitable,
    Callable,
    ClassVar,
    Generic,
    ParamSpec,
    TypeVar,
    overload,
)

from ._utils import (
    HandleStream,
    HandleStreamAsync,
    SetupCall,
    get_fn_args,
    get_metadata,
)
from .call_params import BaseCallParams
from .call_response import BaseCallResponse
from .call_response_chunk import BaseCallResponseChunk
from .dynamic_config import BaseDynamicConfig
from .metadata import Metadata
from .tool import BaseTool

_BaseCallResponseT = TypeVar("_BaseCallResponseT", bound=BaseCallResponse)
_BaseCallResponseChunkT = TypeVar(
    "_BaseCallResponseChunkT", bound=BaseCallResponseChunk
)
_UserMessageParamT = TypeVar("_UserMessageParamT")
_AssistantMessageParamT = TypeVar("_AssistantMessageParamT")
_MessageParamT = TypeVar("_MessageParamT")
_BaseToolT = TypeVar("_BaseToolT", bound=BaseTool)
_BaseCallParamsT = TypeVar("_BaseCallParamsT", bound=BaseCallParams)
_BaseDynamicConfigT = TypeVar("_BaseDynamicConfigT", bound=BaseDynamicConfig)


class BaseStream(
    Generic[
        _BaseCallResponseT,
        _BaseCallResponseChunkT,
        _UserMessageParamT,
        _AssistantMessageParamT,
        _MessageParamT,
        _BaseToolT,
        _BaseDynamicConfigT,
        _BaseCallParamsT,
    ],
    ABC,
):
    """A base class for streaming responses from LLMs."""

    stream: (
        Generator[tuple[_BaseCallResponseChunkT, _BaseToolT | None], None, None]
        | AsyncGenerator[
            tuple[_BaseCallResponseChunkT, _BaseToolT | None],
            None,
        ]
    )
    metadata: Metadata
    tool_types: list[type[_BaseToolT]] | None
    message_param_type: type[_AssistantMessageParamT]
    call_response_type: type[_BaseCallResponseT]
    model: str | None = None
    cost: float | None = None
    prompt_template: str | None = None
    fn_args: dict[str, Any] | None = None
    dynamic_config: _BaseDynamicConfigT
    messages: list[_MessageParamT]
    call_params: _BaseCallParamsT
    user_message_param: _UserMessageParamT | None = None
    message_param: _AssistantMessageParamT

    _provider: ClassVar[str] = "NO PROVIDER"

    def __init__(
        self,
        stream: Generator[tuple[_BaseCallResponseChunkT, _BaseToolT | None], None, None]
        | AsyncGenerator[
            tuple[_BaseCallResponseChunkT, _BaseToolT | None],
            None,
        ],
        *,
        metadata: Metadata,
        tool_types: list[type[_BaseToolT]] | None,
        message_param_type: type[_AssistantMessageParamT],
        call_response_type: type[_BaseCallResponseT],
        model: str,
        cost: float | None,
        prompt_template: str,
        fn_args: dict[str, Any],
        dynamic_config: _BaseDynamicConfigT,
        messages: list[_MessageParamT],
        call_params: _BaseCallParamsT,
        user_message_param: _UserMessageParamT | None,
        provider: str,
    ):
        """Initializes an instance of `BaseStream`."""
        self.stream = stream
        self.metadata = metadata
        self.tool_types = tool_types
        self.message_param_type = message_param_type
        self.call_response_type = call_response_type
        self.model = model
        self.cost = cost
        self.prompt_template = prompt_template
        self.fn_args = fn_args
        self.dynamic_config = dynamic_config
        self.messages = messages
        self.call_params = call_params
        self.user_message_param = user_message_param
        self.provider = provider

    def __iter__(
        self,
    ) -> Generator[tuple[_BaseCallResponseChunkT, _BaseToolT | None], None, None]:
        """Iterator over the stream and stores useful information."""
        assert isinstance(
            self.stream, Generator
        ), "Stream must be a generator for __iter__"
        content, tool_calls = "", []
        for chunk, tool in self.stream:
            content += chunk.content
            if chunk.input_tokens is not None:
                self.input_tokens = chunk.input_tokens
            if chunk.output_tokens is not None:
                self.output_tokens = chunk.output_tokens
            if chunk.model is not None:
                self.model = chunk.model
            if tool:
                tool_calls.append(tool.tool_call)  # type: ignore
            yield chunk, tool
        kwargs: dict[str, Any] = {"role": "assistant"}
        if "message" in self.message_param_type.__annotations__:
            kwargs["message"] = content
        else:
            kwargs["content"] = content
        if tool_calls:
            kwargs["tool_calls"] = tool_calls
        self.message_param = self.message_param_type(**kwargs)

    def __aiter__(
        self,
    ) -> AsyncGenerator[tuple[_BaseCallResponseChunkT, _BaseToolT | None], None]:
        """Iterates over the stream and stores useful information."""

        async def generator():
            assert isinstance(
                self.stream, AsyncGenerator
            ), "Stream must be an async generator for __aiter__"
            content, tool_calls = "", []
            async for chunk, tool in self.stream:
                content += chunk.content
                if chunk.input_tokens is not None:
                    self.input_tokens = chunk.input_tokens
                if chunk.output_tokens is not None:
                    self.output_tokens = chunk.output_tokens
                if chunk.model is not None:
                    self.model = chunk.model
                if tool:
                    tool_calls.append(tool.tool_call)  # type: ignore
                yield chunk, None
            kwargs: dict[str, Any] = {"role": "assistant"}
            if "message" in self.message_param_type.__annotations__:
                kwargs["message"] = content
                if tool_calls:
                    kwargs["tool_calls"] = tool_calls
            else:
                kwargs["content"] = content
                if tool_calls:
                    kwargs["content"] = [kwargs["content"]] + tool_calls
            self.message_param = self.message_param_type(**kwargs)

        return generator()

    def tool_message_params(self, tools_and_outputs):
        """Returns the tool message parameters for tool call results."""
        return self.call_response_type.tool_message_params(tools_and_outputs)


_BaseCallResponseChunkT = TypeVar(
    "_BaseCallResponseChunkT", bound=BaseCallResponseChunk
)
_BaseDynamicConfigT = TypeVar("_BaseDynamicConfigT", bound=BaseDynamicConfig)
_BaseClientT = TypeVar("_BaseClientT", bound=object)
_BaseCallParamsT = TypeVar("_BaseCallParamsT", bound=BaseCallParams)
_ResponseT = TypeVar("_ResponseT")
_ResponseChunkT = TypeVar("_ResponseChunkT")
_P = ParamSpec("_P")


def stream_factory(
    *,
    TCallResponse: type[_BaseCallResponseT],
    TMessageParamType: type[_AssistantMessageParamT],
    TStream: type[BaseStream],
    setup_call: SetupCall[
        _BaseClientT,
        _BaseDynamicConfigT,
        _BaseCallParamsT,
        _ResponseT,
        _ResponseChunkT,
        _BaseToolT,
    ],
    handle_stream: HandleStream[_ResponseChunkT, _BaseCallResponseChunkT, _BaseToolT],
    handle_stream_async: HandleStreamAsync[
        _ResponseChunkT, _BaseCallResponseChunkT, _BaseToolT
    ],
    provider: str,
):
    @overload
    def decorator(
        fn: Callable[_P, _BaseDynamicConfigT],
        model: str,
        tools: list[type[BaseTool] | Callable] | None,
        json_mode: bool,
        client: _BaseClientT | None,
        call_params: _BaseCallParamsT,
    ) -> Callable[_P, TStream]: ...  # pragma: no cover

    @overload
    def decorator(
        fn: Callable[_P, Awaitable[_BaseDynamicConfigT]],
        model: str,
        tools: list[type[BaseTool] | Callable] | None,
        json_mode: bool,
        client: _BaseClientT | None,
        call_params: _BaseCallParamsT,
    ) -> Callable[_P, Awaitable[TStream]]: ...  # pragma: no cover

    def decorator(
        fn: Callable[_P, _BaseDynamicConfigT | Awaitable[_BaseDynamicConfigT]],
        model: str,
        tools: list[type[BaseTool] | Callable] | None,
        json_mode: bool,
        client: _BaseClientT | None,
        call_params: _BaseCallParamsT,
    ) -> Callable[_P, TStream | Awaitable[TStream]]:
        is_async = inspect.iscoroutinefunction(fn)

        @wraps(fn)
        def inner(*args: _P.args, **kwargs: _P.kwargs) -> TStream:
            assert SetupCall.fn_is_sync(fn)
            fn_args = get_fn_args(fn, args, kwargs)
            dynamic_config = fn(*args, **kwargs)
            create, prompt_template, messages, tool_types, call_kwargs = setup_call(
                model=model,
                client=client,
                fn=fn,
                fn_args=fn_args,
                dynamic_config=dynamic_config,
                tools=tools,
                json_mode=json_mode,
                call_params=call_params,
                extract=False,
            )

            def generator():
                for chunk, tool in handle_stream(
                    create(stream=True, **call_kwargs), tool_types
                ):
                    yield chunk, tool

            return TStream(
                generator(),
                metadata=get_metadata(fn, dynamic_config),
                tool_types=tool_types,
                message_param_type=TMessageParamType,
                call_response_type=TCallResponse,
                model=model,
                cost=None,
                prompt_template=prompt_template,
                fn_args=fn_args,
                dynamic_config=dynamic_config,
                messages=messages,
                call_params=call_params,
                user_message_param=messages[-1]
                if messages[-1]["role"] == "user"
                else None,
                provider=provider,
            )

        @wraps(fn)
        async def inner_async(*args: _P.args, **kwargs: _P.kwargs) -> TStream:
            assert SetupCall.fn_is_async(fn)
            fn_args = get_fn_args(fn, args, kwargs)
            dynamic_config = await fn(*args, **kwargs)
            create, prompt_template, messages, tool_types, call_kwargs = setup_call(
                model=model,
                client=client,
                fn=fn,
                fn_args=fn_args,
                dynamic_config=dynamic_config,
                tools=tools,
                json_mode=json_mode,
                call_params=call_params,
                extract=False,
            )

            async def generator():
                async for chunk, tool in handle_stream_async(
                    await create(stream=True, **call_kwargs), tool_types
                ):
                    yield chunk, tool

            return TStream(
                generator(),
                metadata=get_metadata(fn, dynamic_config),
                tool_types=tool_types,
                message_param_type=TMessageParamType,
                call_response_type=TCallResponse,
                model=model,
                cost=None,
                prompt_template=prompt_template,
                fn_args=fn_args,
                dynamic_config=dynamic_config,
                messages=messages,
                call_params=call_params,
                user_message_param=messages[-1]
                if messages[-1]["role"] == "user"
                else None,
                provider=provider,
            )

        return inner_async if is_async else inner

    return decorator
